#include "elisa3-lib.h"
#include "PositionController.h"
#include "NetworkConnection.h"
#include "Utility.h"
#include "MySwisTrackClient.h"

#define ELISA3_DIAM_CM 4.5
#define OBSTACLE_THR 800    // if one the 3 front proximity sensors value > this value then enable onboard obstacle avoidance, disable it otherwise
#define NUM_ROBOTS_MAX 100

// received from robot
unsigned int proxValue[NUM_ROBOTS_MAX][8];
signed int robTheta[NUM_ROBOTS_MAX], robXPos[NUM_ROBOTS_MAX], robYPos[NUM_ROBOTS_MAX];
unsigned int batteryLevel[NUM_ROBOTS_MAX];
int angle[NUM_ROBOTS_MAX];

// sent to robot
int robotAddress[NUM_ROBOTS_MAX];
int initAddress[NUM_ROBOTS_MAX];
int redLed[NUM_ROBOTS_MAX], greenLed[NUM_ROBOTS_MAX], blueLed[NUM_ROBOTS_MAX];
int lspeed[NUM_ROBOTS_MAX], rspeed[NUM_ROBOTS_MAX];

// position control
unsigned int numRobotsToControl=0;
MySwisTrackClient swistrackClient;
unsigned char targetReached[NUM_ROBOTS_MAX];    // 1 if the robot has reached its target piont
unsigned char targetReachedCount=0;     // sum of all robots having reached its target point in this cycle
unsigned char targetNotReachedCount[NUM_ROBOTS_MAX]={0};    // number of consecutive times the robot hasn't reached its target position
unsigned char identifyRobotParticlesFlag=0;
unsigned char allRobsReachDestFirstTime=0;
double pixelToCm = 0;
signed int targetX[NUM_ROBOTS_MAX], targetY[NUM_ROBOTS_MAX];    // current target point
SYSTEMTIME nextStepTime;
FILETIME nextStepTimeF;
ULONGLONG nextStepTime64;
signed int desiredSpeed = 0;
double speedFactor = 0;     // to get a speed proportional to the distance when going to the target
double kRot = 0.15;
double filteredTheta[NUM_ROBOTS_MAX]; // resulting robot orientation (at the moment it is based only on swisstrack info => estimated from delta position)
double robotDiamPixels = 0;
double angleToTarget = 0;
double rotToTarget[NUM_ROBOTS_MAX];
double rotToTargetPrev[NUM_ROBOTS_MAX];
double distToTarget[NUM_ROBOTS_MAX];
double controllerFreq=0;
unsigned char useRobotTheta[NUM_ROBOTS_MAX];
double estOrientDiff[NUM_ROBOTS_MAX];
unsigned int estOrientConfCounter[NUM_ROBOTS_MAX];
signed int robThetaDiff[NUM_ROBOTS_MAX];
signed int robThetaPrev[NUM_ROBOTS_MAX];

// position control parameters
double targetDistThr = 30;      // given in pixels => the target is reached when distance to target <= this value
signed int maxSpeed = 50;
signed int baseSpeed = 15; //maxSpeed/3;
double lowSpeedThr = 0;         // given in cm => when the distance to the target <= this value then the speed is reduced
double highSpeedThr = 100;      // given in cm => when the distance to the target > this value then the speed is maximum
double kRotInit = 0.12;
double rotDiffThr = 50;         // given in degrees; used to avoid changing continuously the direction
double rotMaxThr = 40;          // given in degrees; used to choose when to use the orientation given by the robot instead of the one estimated from deltaX and deltaY given by swisstrack
double lowRotationThr = 20;     // mm/s
double estOrientConfNeed = 5;   // cycles needed before start trusting estimated orientation from swisstrack
double pixelDistSumThr = 10;    // in pixels

// pattern
#define PATTERN_SMALL 0
#define PATTERN_MIDDLE 1
#define PATTERN_BIG 2
unsigned char currPatternSize = PATTERN_SMALL;
unsigned char patternIndex[NUM_ROBOTS_MAX];
 #define NUM_PATTERN_POINTS 4
 unsigned int numTargetPoints = NUM_PATTERN_POINTS;
 // small size pattern
 signed int patternXSmall[NUM_PATTERN_POINTS] = {250,400,400,250};
 signed int patternYSmall[NUM_PATTERN_POINTS] = {200,200,350,350};
 // middle size pattern
 signed int patternXMiddle[NUM_PATTERN_POINTS] = {250,400,400,250};
 signed int patternYMiddle[NUM_PATTERN_POINTS] = {200,200,350,350};
 // big size pattern
 signed int patternXBig[NUM_PATTERN_POINTS] = {250,400,400,250};
 signed int patternYBig[NUM_PATTERN_POINTS] = {200,200,350,350};
unsigned char patternState = 0;
unsigned char maxLedPower = 25;

// network
char server_reply[2000];
int recv_size;

// various
SYSTEMTIME recvTime, currTime, controlFreqUpdateTime, changePatternTime;
FILETIME recvTimeF, currTimeF, controlFreqUpdateTimeF, changePatternTimeF;
ULONGLONG recvTime64, currTime64, controlFreqUpdateTime64, changePatternTime64;
unsigned char printToFileFlag=0;
FILE *logFile;

void printOnTerminal(int id) {
    curPos(0,0);
    printf("Robot: id=%4d, address=%4d, x=%6.2f, y=%6.2f, theta=%6.2f, batt=%3d%%, l=%+3d, r=%+3d\r\n", id, robotAddress[id], swistrackClient.xCenter[id], swistrackClient.yCenter[id], filteredTheta[id], batteryLevel[id], lspeed[id], rspeed[id]);
    printf("Target: x=%4d, y=%4d, dist=%7.2f, rotToTarget=%6.2f\r\n", targetX[id], targetY[id], distToTarget[id], rotToTarget[id]);
    printf("Controller freq =%3.0f Hz\r\n", controllerFreq);
    printf("q=quit, space=reinit\r\n");
}

void printAllOnTerminal() {
    unsigned int id=0;
    curPos(0,0);
    for(id=0; id<numRobotsToControl; id++) {
        printf("Robot: id=%4d, address=%4d, x=%6.2f, y=%6.2f, theta=%6.2f, batt=%3d%%, l=%+3d, r=%+3d\r\n", id, robotAddress[id], swistrackClient.xCenter[id], swistrackClient.yCenter[id], filteredTheta[id], batteryLevel[id], lspeed[id], rspeed[id]);
        printf("Target: x=%4d, y=%4d, dist=%7.2f, rotToTarget=%6.2f\r\n\n", targetX[id], targetY[id], distToTarget[id], rotToTarget[id]);
    }
    printf("Controller freq=%3.0f\r\n", controllerFreq);
    printf("q=quit, space=reinit\r\n");
}

void enableDebugToFile() {
    logFile = fopen("logFile.csv", "w");
    fprintf(logFile, "id,robX,robY,robTheta,targetX,targetY,targetDist,rotToTarget,l,r\n");
    printToFileFlag=1;
}

void disableDebugToFile() {
    printToFileFlag=0;
    fclose(logFile);
}

void printDebugToFile() {
    unsigned int id=0;

    for(id=0; id<numRobotsToControl; id++) {
        fprintf(logFile, "%d,%6.2f,%6.2f,%6.2f,%4d,%4d,%7.2f,%6.2f,%+3d,%+3d\n", id, swistrackClient.xCenter[id], swistrackClient.yCenter[id], swistrackClient.estOrient[id], targetX[id], targetY[id], distToTarget[id], rotToTarget[id], lspeed[id], rspeed[id]);
    }

}

double computePixelToCmFactor() {

    double xMin[NUM_ROBOTS_MAX], xMax[NUM_ROBOTS_MAX];
    double yMin[NUM_ROBOTS_MAX], yMax[NUM_ROBOTS_MAX];
    double xMinFinal, xMaxFinal, yMinFinal, yMaxFinal;
    SYSTEMTIME startTime, currTime;
    FILETIME startTimeF, currTimeF;
    ULONGLONG startTime64, currTime64;
    unsigned int i = 0;

    // turn on the back IR and start pivot rotating the robots
    for(i=0; i<numRobotsToControl; i++) {
        xMin[i] = 2000;
        yMin[i] = 2000;
        xMax[i] = 0;
        yMax[i] = 0;
        setLeftSpeed(robotAddress[i], -20);
        setRightSpeed(robotAddress[i], 20);
        turnOnBackIR(robotAddress[i]);
        waitForUpdate(robotAddress[i], 100000);
    }


    updateTime(&startTime, &startTimeF, &startTime64);

    // while the robots are moving save their maximum and minimum x a y coordinates
    // to eventually get the diameter (we need only one axis, but if we take both we
    // could check that the measure is correct...)
    while(1) {
        recv_size=ReceiveTCP(server_reply);
        if(recv_size > 0) {
            swistrackClient.processData(server_reply, recv_size);
            for(i=0; i<numRobotsToControl; i++) {
                if(xMin[i] > swistrackClient.xPos[i] && swistrackClient.xPos[i] > 0) {
                    xMin[i] = swistrackClient.xPos[i];
                    //printf("xMin=%f\n", xMin);
                }
                if(xMax[i] < swistrackClient.xPos[i]) {
                    xMax[i] = swistrackClient.xPos[i];
                    //printf("xMax=%f\n", xMax);
                }
                if(yMin[i] > swistrackClient.yPos[i] && swistrackClient.yPos[i] > 0) {
                    yMin[i] = swistrackClient.yPos[i];
                    //printf("yMin=%f\n", yMin);
                }
                if(yMax[i] < swistrackClient.yPos[i]) {
                    yMax[i] = swistrackClient.yPos[i];
                    //printf("yMax=%f\n", yMax);
                }
            }
        }
        updateTime(&currTime, &currTimeF, &currTime64);
        if(diffTimeMilliseconds(startTime64, currTime64) > 2500) { // 2.5 seconds
            break;
        }
    }

    // stop the robots
    for(i=0; i<numRobotsToControl; i++) {
        setLeftSpeed(robotAddress[i], 0);
        setRightSpeed(robotAddress[i], 0);
        waitForUpdate(robotAddress[i], 100000);
    }

    // take the first valid measure (only one robot is enough...)
    for(i=0; i<numRobotsToControl; i++) {
        if((xMax[i]-xMin[i])>5) {   // avoid to take noise as the diameter measure
            xMaxFinal = xMax[i];
            xMinFinal = xMin[i];
            yMaxFinal = yMax[i];
            yMinFinal = yMin[i];
            break;
        }
    }

    printf("delta x = %f\n", xMaxFinal - xMinFinal);
    printf("delta y = %f\n", yMaxFinal - yMinFinal);
    // we can check delta x is more or less equal to delta y to verify that the process was well performed

    robotDiamPixels = (xMaxFinal-xMinFinal);
    printf("pixel to cm factor = %f\n", ELISA3_DIAM_CM/robotDiamPixels);
    return ELISA3_DIAM_CM/robotDiamPixels;

}

int identifyRobotParticles() {

    // move the robots one at a time and see what particle id is moving from swistrack, then couple the
    // robot address with that particle id

    int error = 0;
    double xMin[NUM_ROBOTS_MAX], xMax[NUM_ROBOTS_MAX];
    SYSTEMTIME startTime, currTime;
    FILETIME startTimeF, currTimeF;
    ULONGLONG startTime64, currTime64;
    unsigned int i = 0, j = 0;
    signed int particleOfRobot[NUM_ROBOTS_MAX];
    int tempAddress[NUM_ROBOTS_MAX];

    for(i=0; i<numRobotsToControl; i++) {
        xMin[i] = 2000;
        xMax[i] = 0;
        particleOfRobot[i] = -1;
        tempAddress[i] = 0;
        robotAddress[i] = initAddress[i];
        turnOnBackIR(robotAddress[i]);
        disableObstacleAvoidance(robotAddress[i]);
        setLeftSpeed(robotAddress[i], 0);
        setRightSpeed(robotAddress[i], 0);
        waitForUpdate(robotAddress[i], 100000);
        printf("robot addr=%d, particle id=%d\n", robotAddress[i], i);
    }

    // wait for all robots to be stopped
    updateTime(&startTime, &startTimeF, &startTime64);
    while(1) {
        recv_size=ReceiveTCP(server_reply);
        if(recv_size > 0) {
            swistrackClient.processData(server_reply, recv_size);
        }
        updateTime(&currTime, &currTimeF, &currTime64);
        if(diffTimeMilliseconds(startTime64, currTime64) > 500) { // 0.5 seconds
            break;
        }
    }

    for(j=0; j<numRobotsToControl; j++) {
        printf("robot %d\n", j);

        // let move the robot to be identified (pivot rotation)
        setLeftSpeed(robotAddress[j], -20);
        setRightSpeed(robotAddress[j], 20);
        waitForUpdate(robotAddress[j], 100000);

        // get maximum and minimum x position during rotation
        updateTime(&startTime, &startTimeF, &startTime64);
        while(1) {
            recv_size=ReceiveTCP(server_reply);
            if(recv_size > 0) {

                swistrackClient.processData(server_reply, recv_size);

                for(i=0; i<numRobotsToControl; i++) {
                    if(xMin[i] > swistrackClient.xPos[i] && swistrackClient.xPos[i] > 0) {
                        xMin[i] = swistrackClient.xPos[i];
                        //printf("xMin[%d]=%f\n", i, xMin[i]);
                    }
                    if(xMax[i] < swistrackClient.xPos[i]) {
                        xMax[i] = swistrackClient.xPos[i];
                        //printf("xMax[%d]=%f\n", i, xMax[i]);
                    }
                }
            }
            updateTime(&currTime, &currTimeF, &currTime64);
            if(diffTimeMilliseconds(startTime64, currTime64) > 1500) { // 1.5 seconds
                break;
            }
        }

        // stop the robot
        setLeftSpeed(robotAddress[j], 0);
        setRightSpeed(robotAddress[j], 0);
        waitForUpdate(robotAddress[j], 100000);

        // the robot is still moving a little so flush the position data before passing to the next robot
        // otherwise there could be some wrong identification
        updateTime(&startTime, &startTimeF, &startTime64);
        while(1) {
            recv_size=ReceiveTCP(server_reply);
            if(recv_size > 0) {
                swistrackClient.processData(server_reply, recv_size);
            }
            updateTime(&currTime, &currTimeF, &currTime64);
            if(diffTimeMilliseconds(startTime64, currTime64) > 500) { // 0.5 seconds
                break;
            }
        }

        for(i=0; i<numRobotsToControl; i++) {
            if((xMax[i]-xMin[i])>15) {  // be sure that it is actually a robot and not noise
                printf("particle of robot %d = %d (diff=%f)\n", j, i, xMax[i]-xMin[i]);
                particleOfRobot[j] = i;
            }
            xMax[i] = 0;
            xMin[i] = 2000;
        }

    }

    // change robot address sequence to reflect partciles id
    for(i=0; i<numRobotsToControl; i++) {
        if(particleOfRobot[i] >= 0) {
            tempAddress[particleOfRobot[i]] = robotAddress[i];
        }
    }

    for(i=0; i<numRobotsToControl; i++) {
        robotAddress[i] = tempAddress[i];
        printf("robot addr=%d, particle id=%d\n", robotAddress[i], i);
        if(robotAddress[i] == 0) {
            error = 1;
        }
    }

    setRobotAddresses(robotAddress, numRobotsToControl); // here the robots state is reset (IR turned off, speed to zero)

    for(i=0; i<numRobotsToControl; i++) {
        //setLeftSpeed(robotAddress[i], 0);
        //setRightSpeed(robotAddress[i], 0);
        turnOnBackIR(robotAddress[i]);
        waitForUpdate(robotAddress[i], 100000);
        allRobsReachDestFirstTime = 0;
        targetNotReachedCount[i] = 0;
    }

    allRobsReachDestFirstTime = 0;

    return error;

}

void startPositionController(int *robotAddr, int numRobots) {

    unsigned int i=0;

    // open the connection with Swistrack to get tracking informations
    OpenTCPConnection(3000, "127.0.0.1");

    numRobotsToControl = numRobots;
    for(i=0; i<numRobotsToControl; i++) {
        robotAddress[i] = robotAddr[i];
        initAddress[i] = robotAddr[i];
    }

    while(identifyRobotParticles());

    pixelToCm = computePixelToCmFactor();   // cm = pixel*pixelToCm

    swistrackClient.setNumParticles(numRobots);
    swistrackClient.setRadiusPixels(robotDiamPixels/2);
    printf("radius=%f\n", swistrackClient.radiusPixels);

    lowSpeedThr = lowSpeedThr/pixelToCm;    // cm to pixels
    highSpeedThr = highSpeedThr/pixelToCm;  // cm to pixels
    printf("low speed thr=%f\n", lowSpeedThr);
    printf("high speed thr=%f\n", highSpeedThr);

    speedFactor = (maxSpeed-baseSpeed)/(highSpeedThr-lowSpeedThr);
    printf("speed factor = %f\n", speedFactor);

    for(i=0; i<numRobotsToControl; i++) {
        targetReached[i]=0;
        patternIndex[i] = i;    // set initial position of the robots (point on the pattern)
        targetX[i] = patternXSmall[patternIndex[i]];
        targetY[i] = patternYSmall[patternIndex[i]];
    }



    updateTime(&nextStepTime, &nextStepTimeF, &nextStepTime64);
    updateTime(&recvTime, &recvTimeF, &recvTime64);
    updateTime(&controlFreqUpdateTime, &controlFreqUpdateTimeF, &controlFreqUpdateTime64);
    updateTime(&changePatternTime, &changePatternTimeF, &changePatternTime64);

    clearTerminal();

}

void stopPositionController() {
    CloseTCPConnection();
}

void getCenterOfAdvanced(unsigned int robotIndex) {

    // calculate the orientation variations both from swistrack data and odometry data
    estOrientDiff[robotIndex] = swistrackClient.estOrient[robotIndex] - swistrackClient.estOrientPrev[robotIndex];
    if(estOrientDiff[robotIndex] < -180) {
        estOrientDiff[robotIndex] += 360;
    }
    if(estOrientDiff[robotIndex] > 180) {
        estOrientDiff[robotIndex] -= 360;
    }
    robThetaDiff[robotIndex] = robTheta[robotIndex] - robThetaPrev[robotIndex];

//*****************************************************************************************************************************************//
// The first version decides to use either swistrack or odometry based on the difference between the current orientations from both.
// If the variation gis too big than probably the orientation given by swistrack isn't correct, thus use odometry data.

//    if(abs(abs(estOrientDiff[robotIndex]) - abs(robThetaDiff[robotIndex])) > rotMaxThr) {
//        useRobotTheta[robotIndex] = 1;
//        filteredTheta[robotIndex] += robThetaDiff[robotIndex];
//        filteredTheta[robotIndex] = (float)(((signed int)filteredTheta[robotIndex])%360);
//        if(filteredTheta[robotIndex] < 0) {
//            filteredTheta[robotIndex] += 360.0;
//        }
//    } else {
//        filteredTheta[robotIndex] = swistrackClient.estOrient[robotIndex];
//        useRobotTheta[robotIndex] = 0;
//    }

// end first version
//*****************************************************************************************************************************************//

//*****************************************************************************************************************************************//
// The second version uses the difference between the right and left speed to give an estimation of the current rotation speed.
// When the rotation speed is big doesn't trust estimated angle from swistrack.

  // the estimated orientation given from delta position (swisstrack data) is valid only after the
  // robot is goiong more or less straight for a while
  if(abs(rspeed[robotIndex]-lspeed[robotIndex])<lowRotationThr && !(rspeed[robotIndex]==0 && lspeed[robotIndex]==0)) {  // rotation speed is low and robot not stopped
      if(estOrientConfCounter[robotIndex] < estOrientConfNeed) {
          estOrientConfCounter[robotIndex]++;                                                   // increase the confidence to estimated angle from swistrack at low rotation speed
      } else {
          estOrientConfCounter[robotIndex] = estOrientConfNeed;
      }
  } else {
      estOrientConfCounter[robotIndex] = 0;
      swistrackClient.resetPixelDistances(robotIndex);
  }

  // the robot need to travel a certain distance for the estimated orientation given from delta position
  // (swisstrack data) to be valid; if the robot moves at low speed we cannot trust the resulting estimated
  // orientation
  swistrackClient.sumPixelDistances(robotIndex);

  if(estOrientConfCounter[robotIndex]>=estOrientConfNeed && swistrackClient.pixelDistSum[robotIndex]>=pixelDistSumThr && rspeed[robotIndex]!=0 && lspeed[robotIndex]!=0) { // when the confidence is enough switch to swistrack angle
      useRobotTheta[robotIndex] = 0;
      filteredTheta[robotIndex] = swistrackClient.estOrient[robotIndex];
  } else {
      useRobotTheta[robotIndex] = 1;
      filteredTheta[robotIndex] += robThetaDiff[robotIndex];
      filteredTheta[robotIndex] = (float)(((signed int)filteredTheta[robotIndex])%360);
      if(filteredTheta[robotIndex] < 0) {
          filteredTheta[robotIndex] += 360.0;
      }
  }
// end second version
//*****************************************************************************************************************************************//

    // compute the real center of the robot based on diameter size in pixels and current orientation of the robot
    swistrackClient.xCenter[robotIndex] = swistrackClient.xPos[robotIndex] + (robotDiamPixels/2)*cos(filteredTheta[robotIndex]/180*M_PI);
    swistrackClient.yCenter[robotIndex] = swistrackClient.yPos[robotIndex] - (robotDiamPixels/2)*sin(filteredTheta[robotIndex]/180*M_PI);
}

void getCenterOfBasic(unsigned int robotIndex) {

    //filteredTheta[robotIndex] = swistrackClient.estOrient[robotIndex];   // use always the angle estimated from swistrack at the moment
    angle[robotIndex] = getVerticalAngle(robotAddress[robotIndex]);
    filteredTheta[robotIndex] = angle[robotIndex];   // use angle from accelerometer when in vertical

    // compute the real center of the robot based on diameter size in pixels and current orientation of the robot
    swistrackClient.xCenter[robotIndex] = swistrackClient.xPos[robotIndex] + (robotDiamPixels/2)*cos(filteredTheta[robotIndex]/180*M_PI);
    swistrackClient.yCenter[robotIndex] = swistrackClient.yPos[robotIndex] - (robotDiamPixels/2)*sin(filteredTheta[robotIndex]/180*M_PI);

}

void getRotationToTargetOf(unsigned int robotIndex) {

    //printf("x:%f, y:%f, theta:%f, xDiff:%f, yDiff:%f\n", client.xPos, client.yPos, client.estOrient, client.xDiff, client.yDiff);
    angleToTarget = atan2(-(targetY[robotIndex]-swistrackClient.yCenter[robotIndex]), targetX[robotIndex]-swistrackClient.xCenter[robotIndex])*180/M_PI;
    if(angleToTarget < 0) {
        angleToTarget += 360.0;
    }
    rotToTarget[robotIndex] = angleToTarget-filteredTheta[robotIndex];
    if(rotToTarget[robotIndex] > 180) { // rotate to the nearest direction
        rotToTarget[robotIndex] -= 360;
    }
    if(rotToTarget[robotIndex] < -180) {
        rotToTarget[robotIndex] += 360;
    }

    if(abs(rotToTarget[robotIndex]-rotToTargetPrev[robotIndex]) > (360-rotDiffThr)) {   // avoid changing continuously direction when the robot is oriented in the opposite direction respect to the target
        rotToTarget[robotIndex] = rotToTargetPrev[robotIndex];
    } else {
        rotToTargetPrev[robotIndex] = rotToTarget[robotIndex];
    }

}

void getDistanceToTargetOf(unsigned int robotIndex) {

    distToTarget[robotIndex] = sqrt((targetY[robotIndex]-swistrackClient.yCenter[robotIndex])*(targetY[robotIndex]-swistrackClient.yCenter[robotIndex]) + (targetX[robotIndex]-swistrackClient.xCenter[robotIndex])*(targetX[robotIndex]-swistrackClient.xCenter[robotIndex]));

}

void handleRGBLed(int id) {

    if(patternIndex[id] == 0) {
        if(redLed[id] < maxLedPower) {
            redLed[id]++;
        }
        if(greenLed[id] > 0) {
            greenLed[id]--;
        }
        if(blueLed[id] > 0) {
            blueLed[id]--;
        }
    } else if(patternIndex[id] == 1) {
        if(redLed[id] > 0) {
            redLed[id]--;
        }
        if(greenLed[id] < maxLedPower) {
            greenLed[id]++;
        }
        if(blueLed[id] > 0) {
            blueLed[id]--;
        }
    } else if(patternIndex[id] == 2) {
        if(redLed[id] > 0) {
            redLed[id]--;
        }
        if(greenLed[id] > 0) {
            greenLed[id]--;
        }
        if(blueLed[id] < maxLedPower) {
            blueLed[id]++;
        }
    } else if(patternIndex[id] == 3) {
        if(redLed[id] < maxLedPower) {
            redLed[id]++;
        }
        if(greenLed[id] < maxLedPower) {
            greenLed[id]++;
        }
        if(blueLed[id] > 0) {
            blueLed[id]--;
        }
    }

    setRed(robotAddress[id], redLed[id]);
    setGreen(robotAddress[id], greenLed[id]);
    setBlue(robotAddress[id], blueLed[id]);

}

void positionControlExec() {

    unsigned int i=0;

    recv_size=ReceiveTCP(server_reply);

    if(recv_size > 0) {

        updateTime(&currTime, &currTimeF, &currTime64);

        //printf("packet time: %I64d ms\n", diffTimeMilliseconds(recvTime64, currTime64));
        //updateTime(&recvTime, &recvTimeF, &recvTime64);
        if(diffTimeSeconds(controlFreqUpdateTime64, currTime64) >= 1) {
            controllerFreq = 1000.0/((double)diffTimeMilliseconds(controlFreqUpdateTime64, currTime64)/(double)swistrackClient.getPacketCount(0));  // the frequency for all robots is the same thus request the num of packets for robot id=0
            swistrackClient.resetPacketCount(0);
            updateTime(&controlFreqUpdateTime, &controlFreqUpdateTimeF, &controlFreqUpdateTime64);
        }

        swistrackClient.processData(server_reply, recv_size);

        targetReachedCount = 0;
        for(i=0; i<numRobotsToControl; i++) {
            targetReachedCount += targetReached[i];
        }

        if(targetReachedCount == numRobotsToControl) {
            allRobsReachDestFirstTime = 1;
        }

        // if no robots reach its target within the timeout (4 seconds), or all the robots reach its targets then pass to next point
        //if( (diffTimeSeconds(nextStepTime64, currTime64) >= 4) || (targetReachedCount>=numRobotsToControl && allRobsReachDestFirstTime>=1)) {
        // alway stop for a while before passing to the next step in order to save energy and reduce continuous noise (especially for demonstrations)
        if(diffTimeSeconds(nextStepTime64, currTime64) >= 5) {
            // stop all the robots just in case there is a timeout
            for(i=0; i<numRobotsToControl; i++) {
                setLeftSpeed(robotAddress[i], 0);
                setRightSpeed(robotAddress[i], 0);
                waitForUpdate(robotAddress[i], 100000);
            }

            // if a robot doesn't reach its target for more than 4 times consecutively it means that there is a problem
            // (probably some particle id were exchanged). Do again an identification in this case.
            for(i=0; i<numRobotsToControl; i++) {
                if(targetReached[i] == 1) {
                    targetNotReachedCount[i] = 0;
                } else {
                    targetNotReachedCount[i]++;
                    if(targetNotReachedCount[i] > 4) {
                        identifyRobotParticlesFlag = 1;
                        break;
                    }
                }
            }
            if(identifyRobotParticlesFlag == 1) {
                identifyRobotParticlesFlag = 0;
                while(identifyRobotParticles());
                return;
            }

             // set new target point of the current pattern
            for(i=0; i<numRobotsToControl; i++) {
                if(patternIndex[i] < (numTargetPoints-1)) {
                    patternIndex[i]++;
                } else {
                    patternIndex[i] = 0;
                }
                if(currPatternSize == PATTERN_SMALL) {
                    targetX[i] = patternXSmall[patternIndex[i]];
                    targetY[i] = patternYSmall[patternIndex[i]];
                } else if(currPatternSize == PATTERN_MIDDLE) {
                    targetX[i] = patternXMiddle[patternIndex[i]];
                    targetY[i] = patternYMiddle[patternIndex[i]];
                } else if(currPatternSize == PATTERN_BIG) {
                    targetX[i] = patternXBig[patternIndex[i]];
                    targetY[i] = patternYBig[patternIndex[i]];
                }
                targetReached[i]=0;
            }

            updateTime(&nextStepTime, &nextStepTimeF, &nextStepTime64);

        }

        // change pattern size every 6 seconds
        if(diffTimeSeconds(changePatternTime64, currTime64) >= 6) {
            switch(patternState) {
                case 0:
                    currPatternSize = PATTERN_MIDDLE;
                    patternState = 1;
                    break;

                case 1:
                    currPatternSize = PATTERN_BIG;
                    patternState = 2;
                    break;

                case 2:
                    currPatternSize = PATTERN_MIDDLE;
                    patternState = 3;
                    break;

                case 3:
                    currPatternSize = PATTERN_SMALL;
                    patternState = 0;
                    break;
            }
            updateTime(&changePatternTime, &changePatternTimeF, &changePatternTime64);
        }

        // update robots info
        for(i=0; i<numRobotsToControl; i++) {
            getCenterOfBasic(i);
            getRotationToTargetOf(i);   // get the rotation needed to orient the robot to the target
            getDistanceToTargetOf(i);
        }

        // set new speed based on current robots info (and update leds color)
        for(i=0; i<numRobotsToControl; i++) {

            handleRGBLed(i);

            // desired speed proportional to the distance to the target point
            desiredSpeed = baseSpeed + (distToTarget[i]-lowSpeedThr)*speedFactor;
            if(desiredSpeed > maxSpeed) {
                desiredSpeed = maxSpeed;
            }
            if(desiredSpeed < baseSpeed) {
                desiredSpeed = baseSpeed;
            }

            if(distToTarget[i] <= targetDistThr) {  // when near the target stop the robot
                targetReached[i]=1;
                lspeed[i] = 0;
                rspeed[i] = 0;
            } else {    // otherwise adapt the speed
                //kRot = kRotInit + ((float)desiredSpeed/1000.0);
                //kRot = kRotInit + ((float)(desiredSpeed*desiredSpeed)/30000.0);
                kRot = kRotInit + kRotInit*((float)(desiredSpeed-baseSpeed)/(float)(maxSpeed-baseSpeed));
                // speed = speed proportional to distance - rotation + base speed*cos
                // the cos helps rotating faster when the robot is oriented towards the opposite direction respect to the target
                lspeed[i] = desiredSpeed + baseSpeed*cos(rotToTarget[i]/180*M_PI) - (signed int)(kRot*rotToTarget[i]);
                //if(lspeed[0] < 0) {
                //    lspeed[0] = 0;
                //}
                if(lspeed[i] > 100) {
                    lspeed[i] = 100;
                }
                if(lspeed[i] < -100) {
                    lspeed[i] = -100;
                }
                rspeed[i] = desiredSpeed + baseSpeed*cos(rotToTarget[i]/180*M_PI) + (signed int)(kRot*rotToTarget[i]);
                //if(rspeed[0] < 0) {
                //   rspeed[0] = 0;
                //}
                if(rspeed[i] > 100) {
                    rspeed[i] = 100;
                }
                if(rspeed[i] < -100) {
                    rspeed[i] = -100;
                }
            }
            setLeftSpeed(robotAddress[i], lspeed[i]);
            setRightSpeed(robotAddress[i], rspeed[i]);
//            waitForUpdate(robotAddress[i], 100000);// do not wait for the speed to be transmitted otherwise the controller will be delayed

            getAllProximity(robotAddress[i], &proxValue[i][0]);
            batteryLevel[i] = getBatteryPercent(robotAddress[i]);
            robThetaPrev[i] = robTheta[i];
            robTheta[i] = getOdomTheta(robotAddress[i]);

            // enable obstacle avoidance only when one the 3 front proximity sensors detect an obstalce, otherwise disable it.
            // This is to smooth the robots movement.
            if(proxValue[i][0]>=OBSTACLE_THR || proxValue[i][1]>=OBSTACLE_THR || proxValue[i][7]>=OBSTACLE_THR) {
                enableObstacleAvoidance(robotAddress[i]);
            } else {
                disableObstacleAvoidance(robotAddress[i]);
            }
//            waitForUpdate(robotAddress[i], 100000);   // do not wait for the enabling/disabling of the obstacle avoidance otherwise the controller will be delayed

        }

        if(printToFileFlag == 1) {
            printDebugToFile();
        }
        memset(server_reply, 0x00, 300);
    }

}

